@@ -1,14 +1,20 @@
-from .. import dspec
-import numpy as np, random
-import pytest
-from pyuvdata import UVData
-from ..data import DATA_PATH
 import os
-import scipy.signal.windows as windows
+import random
 import warnings
+
+import numpy as np
+import pytest
+import scipy.signal.windows as windows
+from pyuvdata import UVData
+from scipy.special import eval_legendre
+
+from .. import dspec
+from ..data import DATA_PATH
+
 random.seed(0)
 import copy
 
+
 def test_wedge_width():
     # Test boundaries of delay bins
     assert dspec.wedge_width(0, .01, 10) == (1,10)
@@ -39,8 +45,8 @@ def test_delay_filter_dims():
 def test_delay_filter_1D():
     NCHAN = 128
     TOL = 1e-6
-    data = np.ones(NCHAN, dtype=np.complex)
-    wgts = .5*np.ones(NCHAN, dtype=np.complex)
+    data = np.ones(NCHAN, dtype=complex)
+    wgts = .5*np.ones(NCHAN, dtype=complex)
     dmdl, dres, info = dspec.delay_filter(data, wgts, 0., .1/NCHAN, tol=TOL)
     np.testing.assert_allclose(data, dmdl, atol=NCHAN*TOL)
     np.testing.assert_allclose(dres, np.zeros_like(dres), atol=NCHAN*TOL)
@@ -63,12 +69,21 @@ def test_delay_filter_1D():
     dmdl, dres, info = dspec.delay_filter(data, wgts, 0., .1/NCHAN, tol=TOL, skip_wgt=0.5, mode='clean')
     assert info['status']['axis_1'][0] == 'skipped'
 
+def test_are_wgts_binary():
+    # Binary weights array should return True
+    wgts = np.random.choice([0, 1], size=100)
+    assert dspec._are_wgts_binary(wgts)
+
+    # Random weights array should return False
+    wgts = np.random.uniform(0, 1, size=100)
+    assert not dspec._are_wgts_binary(wgts)
+
 def test_delay_filter_2D():
     NCHAN = 128
     NTIMES = 10
     TOL = 1e-6
-    data = np.ones((NTIMES, NCHAN), dtype=np.complex)
-    wgts = np.ones((NTIMES, NCHAN), dtype=np.complex)
+    data = np.ones((NTIMES, NCHAN), dtype=complex)
+    wgts = np.ones((NTIMES, NCHAN), dtype=complex)
     dmdl, dres, info = dspec.delay_filter(data, wgts, 0., .1/NCHAN, tol=TOL)
     np.testing.assert_allclose(data, dmdl, atol=NCHAN*TOL)
     np.testing.assert_allclose(dres, np.zeros_like(dres), atol=NCHAN*TOL)
@@ -106,8 +121,8 @@ def test_delay_filter_leastsq():
     NCHAN = 128
     NTIMES = 10
     TOL = 1e-7
-    data = np.ones((NTIMES, NCHAN), dtype=np.complex)
-    flags = np.zeros((NTIMES, NCHAN), dtype=np.bool)
+    data = np.ones((NTIMES, NCHAN), dtype=complex)
+    flags = np.zeros((NTIMES, NCHAN), dtype=bool)
     sigma = 0.1 # Noise level (not important here)
 
     # Fourier coeffs for input data, ordered from (-nmax, nmax)
@@ -162,8 +177,8 @@ def test_skip_wgt():
     NCHAN = 128
     NTIMES = 10
     TOL = 1e-6
-    data = np.ones((NTIMES, NCHAN), dtype=np.complex)
-    wgts = np.ones((NTIMES, NCHAN), dtype=np.complex)
+    data = np.ones((NTIMES, NCHAN), dtype=complex)
+    wgts = np.ones((NTIMES, NCHAN), dtype=complex)
     wgts[0, 0:-4] = 0
     dmdl, dres, info = dspec.delay_filter(data, wgts, 0., .1/NCHAN, tol=TOL, skip_wgt=.1)
     np.testing.assert_allclose(data[1:,:], dmdl[1:,:], atol=NCHAN*TOL)
@@ -224,6 +239,52 @@ def test_dft_operator():
     fop1 = dspec.dft_operator(freqs, 0., 1e-6, fundamental_period=200*1e5)
     np.testing.assert_allclose(fop1, y1)
 
+def test_normalized_legendre():
+    kmax = 5
+    NF = 100
+    x = np.linspace(-0.75, 0.75, NF)
+
+    # Check output matches scipy result
+    polynomials = dspec._normalized_legendre(x, kmax)
+    true_polys = np.transpose([np.sqrt(i + 0.5) * eval_legendre(i, x) for i in range(kmax)])
+    np.testing.assert_allclose(polynomials, true_polys)
+
+def test_calculate_amat():
+    """
+    """
+    kmax = 100
+    c = 20
+    eigenvals, eigenvecs = dspec._calculate_amat(kmax, c)
+    # Check that eigenvalues and eigenvectors are the correct size
+    assert eigenvals.shape[0] == kmax
+    assert eigenvecs.shape == (kmax, kmax)
+    # Check orthogonality
+    np.testing.assert_allclose(np.diag(eigenvecs.T @ eigenvecs), np.ones(kmax))
+
+def test_pswf_operator():
+    #test that an error is thrown when we specify more then one
+    #termination method.
+    NF = 100
+    DF = 100e3
+    freqs = np.arange(-NF/2, NF/2)*DF + 150e6
+    pytest.raises(ValueError, dspec.pswf_operator, x=freqs, filter_centers=[0.], filter_half_widths=[1e-6], nterms=[5], eigenval_cutoff=[1e-9])
+    #now calculate PSWF operator matrices using different cutoff criteria. The columns
+    #should be the same up to the minimum number of columns of the three techniques.
+    amat1, ncol1 = dspec.pswf_operator(freqs, [0.], [100e-9], eigenval_cutoff=[1e-9])
+    amat2, ncol2 = dspec.pswf_operator(freqs, [0.], [100e-9], nterms=[50])
+    ncols = [ncol1, ncol2]
+    ncolmin = np.min(ncols)
+    #check that all columns of matrices obtained with different methods
+    #of cutoff are identical.
+    for m in range(ncolmin):
+        np.testing.assert_allclose(np.abs(amat1[:,m]), np.abs(amat2[:,m]))
+    # Compute partial pswf w/ xmin and xmax and show it gives more terms than amat1
+    amat3, ncol3 = dspec.pswf_operator(freqs, [0.], [100e-9], eigenval_cutoff=[1e-9], xmin=freqs.min() / 2, xmax=freqs.max() * 2)
+    assert sum(ncol3) > sum(ncol1)
+    # Compute partial pswf w/ xmin and xmax
+    amat1, ncol1 = dspec.pswf_operator(freqs, [0.], [100e-9], eigenval_cutoff=[1e-9])
+    amat2, ncol2 = dspec.pswf_operator(freqs, [0., 100e-9], [100e-9, 100e-9], eigenval_cutoff=[1e-9, 1e-9])
+    assert sum(ncol2) == 2 * sum(ncol1)
 
 def test_dpss_operator():
     #test that an error is thrown when we specify more then one
@@ -273,14 +334,18 @@ def test_fit_solution_matrix():
     wmat = np.diag(wgts)
     fitmat_dft = dspec.fit_solution_matrix(wmat, amat_dft)
     fitmat_dpss = dspec.fit_solution_matrix(wmat, amat_dpss)
+    fitmat_dpss_diag = dspec.fit_solution_matrix(wgts, amat_dpss)
     interp_dft = amat_dft @ fitmat_dft @ dw
     interp_dpss = amat_dpss @ fitmat_dpss @ dw
+    interp_dpss_diag = amat_dpss @ fitmat_dpss_diag @ dw
     #DFT interpolation is meh, so we keep our standards low.
     #DFT interpolation matrices are poorly conditioned so that's also
     #Downer.
     assert np.all(np.isclose(interp_dft, data, atol=1e-2))
     #DPSS interpolation is clutch. We can make our standards high.
     assert np.all(np.isclose(interp_dpss, data, atol=1e-6))
+    # DPSS interpolation using diagonal wgts grid
+    assert np.all(np.isclose(interp_dpss_diag, data, atol=1e-6))
     #Check Raising of ValueErrors.
     amat_dft_pc = dspec.dft_operator(fs, [0.], [4. / 50.], fundamental_period=200.)
     with warnings.catch_warnings(record=True) as w:
@@ -524,7 +589,7 @@ def test_vis_filter():
     dfr, ddly = frs[1] - frs[0], dlys[1] - dlys[0]
     d = 200 * np.exp(-2j*np.pi*times[:, None]*(frs[2]+dfr/4) - 2j*np.pi*freqs[None, :]*(dlys[2]+ddly/4)/1e9)
     d += 50 * np.exp(-2j*np.pi*times[:, None]*(frs[20]) - 2j*np.pi*freqs[None, :]*(dlys[20])/1e9)
-    d += 10 * ((np.random.normal(0, 1, uvd.Nfreqs * uvd.Ntimes).astype(np.complex) \
+    d += 10 * ((np.random.normal(0, 1, uvd.Nfreqs * uvd.Ntimes).astype(complex) \
          + 1j * np.random.normal(0, 1, uvd.Nfreqs * uvd.Ntimes)).reshape(uvd.Ntimes, uvd.Nfreqs))
 
     def get_snr(clean, fftax=1, avgax=0, modes=[2, 20]):
@@ -537,12 +602,12 @@ def test_vis_filter():
     freq_snr1, freq_snr2 = get_snr(d, fftax=1, avgax=0, modes=[2, 20])
     time_snr1, time_snr2 = get_snr(d, fftax=0, avgax=1, modes=[2, 20])
     # simulate some flags
-    f = np.zeros_like(d, dtype=np.bool)
+    f = np.zeros_like(d, dtype=bool)
     d[:, 20:22] += 1e3
     f[:, 20:22] = True
     d[20, :] += 1e3
     f[20, :] = True
-    w = (~f).astype(np.float)
+    w = (~f).astype(float)
     bl_len = 70.0 / 2.99e8
 
     # try passing skip_wgt
@@ -703,12 +768,12 @@ def test_fourier_filter():
     time_snr1, time_snr2 = get_snr(d, fftax=0, avgax=1, modes=[2, 20])
 
     # simulate some flags
-    f = np.zeros_like(d, dtype=np.bool)
+    f = np.zeros_like(d, dtype=bool)
     d[:, 20:22] += 1e3
     f[:, 20:22] = True
     d[20, :] += 1e3
     f[20, :] = True
-    w = (~f).astype(np.float)
+    w = (~f).astype(float)
     bl_len = dlys[nf//2+4]
     fr_len = frs[ntimes//2+4]
     # dpss filtering
@@ -741,6 +806,9 @@ def test_fourier_filter():
     mdl4, res4, info4 = dspec.fourier_filter(freqs, d, w, [0.], [bl_len],
                                              mode='clean', **clean_options1)
 
+    mdl5, res5, info5 = dspec.fourier_filter(freqs, d, w, [0.], [bl_len], suppression_factors=[0.],
+                                             mode='dpss_solve', **dpss_options1)
+
 
     clean_options_typo = {'tol':1e-9, 'maxiter':100, 'filt2d_mode':'rect',
                     'edgecut_low':0, 'edgecut_hi':0, 'add_clean_residual':False,
@@ -751,6 +819,8 @@ def test_fourier_filter():
 
     assert np.all(np.isclose(mdl3, mdl4, atol=1e-6))
     assert np.all(np.isclose(res3, res4, atol=1e-6))
+    assert np.all(np.isclose(mdl1, mdl5, atol=1e-6))
+    assert np.all(np.isclose(res1, res5, atol=1e-6))
 
     assert np.all(np.isclose(mdl1, mdl2, atol=1e-6))
     assert np.all(np.isclose(res1, res2))
@@ -946,14 +1016,14 @@ def test_vis_clean():
     dfr, ddly = frs[1] - frs[0], dlys[1] - dlys[0]
     d = 200 * np.exp(-2j*np.pi*times[:, None]*(frs[2]+dfr/4) - 2j*np.pi*freqs[None, :]*(dlys[2]+ddly/4)/1e9)
     d += 50 * np.exp(-2j*np.pi*times[:, None]*(frs[20]) - 2j*np.pi*freqs[None, :]*(dlys[20])/1e9)
-    d += 10 * ((np.random.normal(0, 1, uvd.Nfreqs * uvd.Ntimes).astype(np.complex) \
+    d += 10 * ((np.random.normal(0, 1, uvd.Nfreqs * uvd.Ntimes).astype(complex) \
          + 1j * np.random.normal(0, 1, uvd.Nfreqs * uvd.Ntimes)).reshape(uvd.Ntimes, uvd.Nfreqs))
-    f = np.zeros_like(d, dtype=np.bool)
+    f = np.zeros_like(d, dtype=bool)
     d[:, 20:22] += 1e3
     f[:, 20:22] = True
     d[20, :] += 1e3
     f[20, :] = True
-    w = (~f).astype(np.float)
+    w = (~f).astype(float)
     bl_len = 70.0 / 2.99e8
     # here is a fourier filter implementation of clean
     mdl1, res1, info1 = dspec.fourier_filter(freqs, d, w, [0.], [bl_len],
@@ -1136,6 +1206,47 @@ def test__fit_basis_1d():
 
     assert np.all(np.isclose((mod2+resid2)*wgts, dw, atol=1e-6))
 
+    # Check errors
+    with pytest.raises(ValueError):
+        mod1, resid1, info1 = dspec._fit_basis_1d(fs, dw, np.zeros_like(dw), [0.], [10./3e8], basis_options={'eigenval_cutoff': [1e-12]},
+                                method='undefined', basis='dpss')
+
+    with pytest.raises(ValueError):
+        mod1, resid1, info1 = dspec._fit_basis_1d(fs, dw, np.zeros_like(dw), [0.], [10./3e8], basis_options={'eigenval_cutoff': [1e-12]},
+                                method='solve', basis='undefined')
+
+    # Test with weights grid that is not like mask
+    wgts = np.ones_like(data)
+    wgts += 0.01
+    mod1, resid1, info1 = dspec._fit_basis_1d(fs, dw, wgts, [0.], [5./50.], basis_options=dpss_opts,
+                                    method='leastsq', basis='dpss')
+    mod2, resid2, info2 = dspec._fit_basis_1d(fs, dw, wgts, [0.], [5./50.], basis_options=dpss_opts,
+                                    method='matrix', basis='dpss')
+    mod3, resid3, info3 = dspec._fit_basis_1d(fs, dw, wgts, [0.], [5./50.], basis_options=dpss_opts,
+                                    method='solve', basis='dpss')
+    assert np.allclose(mod1, mod2)
+    assert np.allclose(mod1, mod3)
+
+    with pytest.raises(ValueError):
+        mod1, resid1, info1 = dspec._fit_basis_1d(fs, dw, wgts, [0.], [10./3e8], basis_options={'eigenval_cutoff': [1e-12]},
+                                method='undefined', basis='dpss')
+
+    # Run to trigger skip
+    mod1, _, info1 = dspec._fit_basis_1d(fs, dw, np.zeros_like(dw), [0.], [5/50.], basis_options={'eigenval_cutoff': [1e-12]},
+                                method='solve', basis='dpss')
+    assert info1['skipped']
+    mod1, _, info1 = dspec._fit_basis_1d(fs, dw, np.full_like(dw, np.nan), [0.], [5/50.], basis_options={'eigenval_cutoff': [1e-12]},
+                                method='leastsq', basis='dpss')
+    assert info1['skipped']
+
+    # Trigger cache for solve
+    cache = {}
+    mod1, _, info1 = dspec._fit_basis_1d(fs, dw, wgts, [0.], [5/50.], basis_options={'eigenval_cutoff': [1e-12]},
+                                method='solve', basis='dpss', cache=cache)
+    mod2, _, info2 = dspec._fit_basis_1d(fs, dw, wgts, [0.], [5/50.], basis_options={'eigenval_cutoff': [1e-12]},
+                                method='solve', basis='dpss', cache=cache)
+    assert np.allclose(mod1, mod2)
+
 def test_fit_basis_1d_with_missing_channels():
     fs = np.arange(-50,50)
     #here is some data
